#include <zeno/zeno.h>
#include <zeno/types/PrimitiveObject.h>
#include <zeno/types/PrimitiveUtils.h>
#include <zeno/types/StringObject.h>
#include <zeno/types/NumericObject.h>
#include <zeno/utils/variantswitch.h>
#include <zeno/utils/logger.h>
#include <unordered_map>
#include <cassert>
#include <limits>
#include <queue>
namespace zeno {
namespace {
class myGreator
{
public:
    int operator() (const std::pair<int, float>& p1, const std::pair<int, float>& p2)
    {
        return p1.second > p2.second;
    }
};

void dijkstra(PrimitiveObject *prim, int start, std::vector<float> &dist)
{
    std::priority_queue<std::pair<int, float>, std::vector<std::pair<int,float>>, myGreator> Q;
    dist.resize(prim->verts.size());
    std::vector<std::vector<int>> neigh;
    for(size_t i=0; i<prim->verts.size(); i++)
    {
        dist[i] = -1;
    }
    neigh.resize(prim->verts.size());
    for (int i=0;i<prim->lines.size();i++)
    {
        auto l = prim->lines[i];
        auto v1 = l[0];
        auto v2 = l[1];
        neigh[v1].push_back(v2);
        neigh[v2].push_back(v1);
    }

    dist[start] = 0;
    Q.push(std::make_pair(start, 0.0f));
    while(! Q.empty())
    {
        int u = Q.top().first;
        Q.pop();
        for(auto v:neigh[u])
        {
            float d = distance(prim->verts[u], prim->verts[v]);
            if(dist[v]==-1 || dist[v]>dist[u]+d)
            {
                dist[v] = dist[u] + d;
                Q.push(std::make_pair(v, dist[v]));
            }
        }
    }


}
}
/* AWAK, NIKOLA TESLA'S JOB IS DJ, I.E. DJ-TESLA */
/* THAT EXPLAINS WHY DJ COULD DRIVE THE TESLA CAR */
ZENO_API void primLineDistance(PrimitiveObject *prim, std::string resAttr, int start) {
    if (!prim->verts.size()) return;
    if (!prim->lines.size()) return;
    
    std::vector<float> dist;
    dijkstra(prim, start, dist);
    prim->verts.add_attr<float>(resAttr) = std::move(dist);
    //for(size_t i=0;i<result.size();i++)
    //{
        //result[i] = dist[i];
    //}
    // std::unordered_multimap<int, std::pair<int, float>> neigh;
    // for (int i = 0; i < prim->lines.size(); i++) {
    //     auto line = prim->lines[i];
    //     auto dist = distance(prim->verts[line[0]], prim->verts[line[1]]);
    //     neigh.emplace(line[0], std::pair{line[1], dist});
    //     neigh.emplace(line[1], std::pair{line[0], dist});
    // }

    // auto &result = prim->verts.add_attr<float>(resAttr);
    // std::fill(result.begin(), result.end(), -1);
    // result[start] = 0;

    // std::vector<float> table(prim->verts.size(), -1);
    // {
    //     auto [b, e] = neigh.equal_range(start);
    //     for (auto it = b; it != e; ++it) {
    //         table[it->second.first] = it->second.second;
    //     }
    // }
    // table[start] = -1;

    // for (int i = 1; i < prim->verts.size(); i++) {
    //     float minValue = std::numeric_limits<float>::max();
    //     int minIndex = 0;
    //     for (int j = 0; j < table.size(); j++) {
    //         if (table[j] >= 0 && table[j] < minValue) {
    //             minValue = table[j];
    //             minIndex = j;
    //         }
    //     }
    //     result[minIndex] = minValue;
    //     //printf("%d %f\n", minIndex, minValue);
    //     table[minIndex] = -1;
    //     for (int j = 0; j < table.size(); j++) {
    //         auto [b, e] = neigh.equal_range(minIndex);
    //         auto jit = std::find_if(b, e, [&] (auto const &pa) {
    //             return pa.second.first == j;
    //         });
    //         if (jit != e && result[j] < 0) {
    //             float newDist = result[minIndex] + jit->second.second;
    //             if (table[j] < 0 || newDist < table[j]) {
    //                 table[j] = newDist;
    //             }
    //         }
    //     }
    // }
}

struct PrimitiveLineDistance : zeno::INode {
    virtual void apply() override {
        auto prim = get_input<zeno::PrimitiveObject>("prim");
        primLineDistance(prim.get(),
                         get_input2<std::string>("resAttr"),
                         get_input2<int>("start"));
        set_output("prim", std::move(prim));
    }
};

ZENDEFNODE(PrimitiveLineDistance, {
    {
    {"PrimitiveObject", "prim"},
    {"string", "resAttr", "len"},
    {"int", "start", "0"},
    },
    {
    {"PrimitiveObject", "prim"},
    },
    {
    },
    {"primitive"},
});

}
